#include "xparameters.h"
#include "xbasic_types.h"   /* Xilinx basic types for device drivers        */
#include "xac97.h"
#include "xac97_l.h"
#include "xac97_i.h"
#include "mb_interface.h"

#define TRUE 1
#define FALSE 0
#define printf xil_printf
#define audio_fsl 0 //using these in bread/bwrite commands is causing errors, don't know why
#define fft_fsl 1 //using these in bread/bwrite commands is causing errors, don't know why

#define XFFT_mWrite32(Base, Offset, Value) \
		XIo_Out32(((XIo_Address)((Base)+(Offset))),(Value))

// Performs a general 32-bit read operation from specified address
#define XFFT_mRead32(Base, Offset) \
		XIo_In32((XIo_Address)((Base)+Offset))

// Performs a general 16-bit write operation to specified address
#define XFFT_mWrite16(Base, Offset, Value) \
		XIo_Out16(((XIo_Address)((Base)+(Offset))),(Value))

// Performs a general 16-bit read operation from specified address
#define XFFT_mRead16(Base, Offset) \
		XIo_In16((XIo_Address)((Base)+Offset))

#define SIZE 512

// Globals
Xuint16 audio1[SIZE];
Xuint16 audio2[SIZE];
Xuint32 coeff1[SIZE];
Xuint32 coeff2[SIZE];

void fft(int chan, int scale, int length, int dir)
{
	/* performs forward/inverse fft operation on global arrays */

	Xuint32 nfft;
	Xuint32 scale_sch;
	Xuint32 fwd_inv;
	Xuint32 cmd = 0x00000000;
	int i;

	switch (scale) {
		case 1:
			scale_sch = 0xA0000000;
			break;
		case 2:
			scale_sch = 0xAA000000;
			break;
		case 3:
			scale_sch = 0x55410000;
			break;
		default:
			scale_sch = 0xd5540000; // conservative scaling
			break;
	}

	switch (length) {
		case 128:
			nfft = 0x0000E000;
			break;
		case 256:
			nfft = 0x00001000;
			break;
		case 512:
			nfft = 0x00009000;
			break;
		case 1024:
			nfft = 0x00005000;
			break;
		case 2048:
			nfft = 0x0000D000;
			break;
		case 4096:
			nfft = 0x00003000;
			break;
		case 8192:
			nfft = 0x0000B000;
			break;
		default:
			nfft = 0x00006000; // 64 point FFT
			break;
	}

	if (dir == 1) // forward FFT operation request
		fwd_inv = 0x00000001;
	else
		fwd_inv = 0x00000000;

	cmd = scale_sch | nfft | fwd_inv; //configuration data to write to OPB Data bus

	//write to FSL bus going into FFT
	if (fwd_inv == 1){ // forward FFT audio and dump into coefficients array
		if (chan == 1){
			for (i=0; i<length; i++){
				microblaze_bwrite_datafsl(audio1[i], 1);
			}
			XFFT_mWrite32(XPAR_FFT_0_BASEADDR, 0, cmd);
			for (i=0; i<length; i++){
				microblaze_bread_datafsl(coeff1[i], 1);
			}
		}
		else{
			for (i=0; i<length; i++){
				microblaze_bwrite_datafsl(audio2[i], 1);
			}
			XFFT_mWrite32(XPAR_FFT_0_BASEADDR, 0, cmd);
			for (i=0; i<length; i++){
				microblaze_bread_datafsl(coeff2[i], 1);
			}
		}
	}
	else{//inverse FFT coefficients and dump into audio array
		if (chan == 1){
			for (i=0; i<length; i++){
				microblaze_bwrite_datafsl(coeff1[i], 1);
			}
			XFFT_mWrite32(XPAR_FFT_0_BASEADDR, 0, cmd);
			for (i=0; i<length; i++){
				microblaze_bread_datafsl(audio1[i], 1);
			}
		}
		else{
			for (i=0; i<length; i++){
				microblaze_bwrite_datafsl(coeff2[i], 1);
			}
			XFFT_mWrite32(XPAR_FFT_0_BASEADDR, 0, cmd);
			for (i=0; i<length; i++){
				microblaze_bread_datafsl(audio2[i], 1);
			}
		}
	}

	return;
}

void initialize(int sample_rate)
{
	/* Initialization of AC97 core.  See xac97.c*/

	printf("starting initialization\r\n");
	init_sound(sample_rate);
	printf("completed initialization\r\n");
	playback_enable();
	printf("completed playback enable\r\n");
	record_enable();
	printf("completed record enable\r\n");
	printf("starting program...\r\n");

	return;
}


void recbuffer(int length)
{
	/* Records buffers of audio up to 512.  FSL parameter for this bus is set to 1024 but
	can only smoothly play audio half of the depth parameter.
	To perform larger FFT bursts, loop transfers to/from FFT FSL and AC97 FSL*/
	int i;
	for (i=0;i<length;i++){
		microblaze_bread_datafsl(audio1[i], 0);
		microblaze_bread_datafsl(audio2[i], 0);
	}

	return;
}

void playbuffer(int length, int sel)
{
	/* Plays buffers of audio up to 512.  FSL parameter for this bus is set to 1024 but
	can only smoothly play audio half of the depth parameter.
	To perform larger FFT bursts, loop transfers to/from FFT FSL and AC97 FSL*/

	int i;
	if (sel == 1){
		for (i=0;i<length;i++){
			microblaze_bwrite_datafsl(audio1[i], 0);
			microblaze_bwrite_datafsl(0, 0);
		}
	}
	else if (sel == 2){
		for (i=0;i<length;i++){
			microblaze_bwrite_datafsl(0, 0);
			microblaze_bwrite_datafsl(audio2[i], 0);
		}
	}
	else{
		for (i=0;i<length;i++){
			microblaze_bwrite_datafsl(audio1[i], 0);
			microblaze_bwrite_datafsl(audio2[i], 0);
		}
	}
}


void makenoise(int length, int chan)
{
	/* Generates white noise audio by IFFTing constant power spectrum.  Not fully verified
	or tested*/

	int i;
	if (chan == 1){
		for (i=0; i<length; i++){
			coeff1[i]=0x00000000;
		}
		coeff1[0]=0x55555555;
		fft(chan, 2, length, 0);//(int chan, int scale, int length, int dir)
	}
	else if (chan == 2){
		for (i=0; i<length; i++){
			coeff2[i]=0x22222222;
		}
		fft(chan, 2, length, 0);//(int chan, int scale, int length, int dir)
	}
	else{
		for (i=0; i<length; i++){
			coeff1[i]=0x22222222;
			coeff2[i]=0x22222222;
		}
		fft(1, 2, length, 0);//(int chan, int scale, int length, int dir)
		fft(2, 2, length, 0);//(int chan, int scale, int length, int dir)
	}
	return;
}

void vocode(int length, int result_chan, int scale)
{
	/* Performs vocoding by imposing spectrum of modulator onto the spectrum
	of the carrier (usually sounds best when carrier is frequency rich, like noise).
	Currently will perform vocoding on all channels of the FFT, but typically will 
	do only 8-channel, 16-channels, etc. in commercially available vocoders.  
	Carrier (noise) should be fed in on one of the audio channels simultaneously
	with modulator*/

	int i;
	Xuint16 real;
	Xuint32 img;

	if (result_chan == 1){
		for (i=0; i<length; i++){
			real = ((coeff1[i] & 0x0000ffff) >> scale) * ((coeff2[i] & 0x0000ffff) >> scale);
			real = real & 0x0000ffff;
			img = ((coeff1[i] & 0xffff0000) >> scale) * ((coeff2[i] & 0xffff0000) >> scale);
			
			//gate level
			if (real < 0x00000500)
				real = 0;
			if (img < 0x05000000)
				img = 0;
			coeff1[i] = img | real;
		}
	}
	else{
		for (i=0; i<length; i++){
			real = ((coeff1[i] & 0x0000ffff) >> scale) * ((coeff2[i] & 0x0000ffff) >> scale);
			real = real & 0x0000ffff;
			img = ((coeff1[i] & 0xffff0000) >> scale) * ((coeff2[i] & 0xffff0000) >> scale);

			//gate level
			if (real < 0x00000500)
				real = 0;
			if (img < 0x05000000)
				img = 0;
			coeff2[i] = img | real;
		}
	}
}

void eq(int length, int sel, float bass, float midbass, float midtreb, float treb)
{
	/* Performs EQ by scaling appropriate coefficients. Not fully tested or verified*/

	int i;
	Xuint16 real;
	Xuint32 img;

	if (sel == 1){
		// bass scaling
		for (i=0; i<length/4; i++){
			real = (coeff1[i] & 0x0000ffff) * bass;
			img = (coeff1[i] & 0xffff0000) * bass;
			coeff1[i] = img | real;
		}

		// mid bass scaling
		for (i=length/4; i<length/2; i++){
			real = (coeff1[i] & 0x0000ffff) * midbass;
			img = (coeff1[i] & 0xffff0000) * midbass;
			coeff1[i] = img | real;
		}

		// mid treble scaling
		for (i=length/2; i<3*length/4; i++){
			real = (coeff1[i] & 0x0000ffff) * midtreb;
			img = (coeff1[i] & 0xffff0000) * midtreb;
			coeff1[i] = img | real;
		}

		// treble scaling
		for (i=3*length/4; i<length; i++){
			real = (coeff1[i] & 0x0000ffff) * treb;
			img = (coeff1[i] & 0xffff0000) * treb;
			coeff1[i] = img | real;
		}
	}
	else if (sel == 2){
		// bass scaling
		for (i=0; i<length/4; i++){
			real = (coeff2[i] & 0x0000ffff) * bass;
			img = (coeff2[i] & 0xffff0000) * bass;
			coeff2[i] = img | real;
		}

		// mid bass scaling
		for (i=length/4; i<length/2; i++){
			real = (coeff2[i] & 0x0000ffff) * midbass;
			img = (coeff2[i] & 0xffff0000) * midbass;
			coeff2[i] = img | real;
		}

		// mid treble scaling
		for (i=length/2; i<3*length/4; i++){
			real = (coeff2[i] & 0x0000ffff) * midtreb;
			img = (coeff2[i] & 0xffff0000) * midtreb;
			coeff2[i] = img | real;
		}

		// treble scaling
		for (i=3*length/4; i<length; i++){
			real = (coeff2[i] & 0x0000ffff) * treb;
			img = (coeff2[i] & 0xffff0000) * treb;
			coeff2[i] = img | real;
		}
	}

	return;
}

int main (void)
{
	int i;
	initialize(44100);

	while(1) {
		recbuffer(512); //(int length)
		//makenoise(512, 1);
		fft(1, 2, 512, 1);//(int chan, int scale, int length, int dir)
		fft(2, 2, 512, 1);//(int chan, int scale, int length, int dir)
		//vocode(512, 1, 6);//(int length, int result_chan, int scale)
		//eq(512, 1, 1, 1, 1, 1);//(int length, int sel, float bass, float midbass, float midtreb, float treb)
		fft(1, 2, 512, 0);//(int chan, int scale, int length, int dir)
		fft(2, 2, 512, 0);//(int chan, int scale, int length, int dir)
		playbuffer(512, 2);//(int length, int sel)
	}
	return 0;
}